/*****************************************************************************

NAME:
   main.c -- main sequence and protocol handling for bogofilter

THEORY:
   Both client and server logic for the bogofilter autodaemon.  The reason
for the weird inside-out control structure, hooks piled on hooks, is so
that all of the process-spawning stuff can live in autodaemon.c while all of
the protocol interpretation on both sides lives here.

AUTHOR:
   Eric S. Raymond <esr@thyrsus.com>

******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <limits.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <limits.h>
#include "bogofilter.h"

#define BOGODIR		"/.bogofilter/"
#define GOODFILE        "goodlist"
#define BADFILE		"badlist"

static char goodfile[PATH_MAX], badfile[PATH_MAX];

static int verbose, passthrough;

struct textblock
{
    char		*block;
    struct textblock	*next;
};

static struct textblock textblocks, *textend = &textblocks;

static ship(FILE *infp, int fd)
// Copy from a file pointer to a socket descriptor
{
    char	block[BUFSIZ], *lp;

    while ((lp = fgets(block, sizeof(block), infp)) != NULL)
    {
	if (passthrough)
	{
	    // Also, save the text on a linked list of lines.
	    // Note that we store fixed-length blocks here, not lines.
	    // One very long phsyical line could break up into more
	    // than one of these.
	    textend->block = strdup(block);
	    textend->next = (struct textblock *)malloc(sizeof(struct textblock));
	    textend->next->block = (char *)NULL;
	    textend = textend->next;
	}

	sockputs(block, fd);
    }
}

static char *accept_response(int fd, int chatty)
// Accept and echo a server response, including informational messages
{
    static char	recvline[BUFSIZ];

    for (;;)
    {
	sockgets(fd, recvline, sizeof(recvline));
	if (chatty && recvline[0] == '#' || chatty > 1)
	    fputs(recvline, stdout);
	if (recvline[0] != '#')
	    break;
    }

    return(recvline);
}

static int singleshot(const char *cmd, int fd)
// Send down a single command followed by message text, no response needed
{
    sockputs(fd, cmd, strlen(cmd));
    ship(stdin, fd);
    write(fd, ".\n", 2);
    return(0);
}

static int register_nonspam(int fd) {return singleshot("NONSPAM\n", fd);}
static int spam_to_nonspam(int fd)  {return singleshot("DESPAM\n",  fd);}
static int register_spam(int fd)    {return singleshot("SPAM\n",    fd);}
static int nonspam_to_spam(int fd)  {return singleshot("ENSPAM\n",  fd);}

static int unknown(int fd)
// Query the spamicity of a message
{
    char	*cp;

    sockputs(fd, "UNKNOWN\n", 8);
    ship(stdin, fd);
    write(fd, ".\n", 2);
    cp = accept_response(fd, verbose);

    return(!!strcmp(cp, "SPAM\n"));
}

static int debug(int fd)
// Debug & test frame for the client
{
    char	sendline[BUFSIZ];

#define SHIP(str)	write(fd, str, strlen(str));
    printf("Welcome to the bogofilter debug tester.\n");
    
    for (;;) 
    {
	int	n;
	char	recvline[PATH_MAX], arg1[PATH_MAX];

	fputs("> ", stdout);
	if (fgets(sendline, BUFSIZ, stdin) == NULL)
	    break;

	// Interpretation
	if (sscanf(sendline, "s %s") == 1)
	{
	    FILE	*infp = fopen(arg1, "r");

	    if (infp)
	    {
		printf("Registering %s as spam.\n", arg1);
		SHIP("SPAM\n");
		ship(infp, fd);
		SHIP(".\n");
	    }
	}
	else if (sscanf(sendline, "n %s\n") == 1)
	{
	    FILE	*infp = fopen(arg1, "r");

	    if (infp)
	    {
		printf("Registering %s as nonspam.\n", arg1);
		SHIP("NONSPAM\n");
		ship(infp, fd);
		SHIP(".\n");
	    }
	}
	if (sscanf(sendline, "S %s") == 1)
	{
	    FILE	*infp = fopen(arg1, "r");

	    if (infp)
	    {
		printf("Registering %s as spam.\n", arg1);
		SHIP("ENSPAM\n");
		ship(infp, fd);
		SHIP(".\n");
	    }
	}
	else if (sscanf(sendline, "N %s\n") == 1)
	{
	    FILE	*infp = fopen(arg1, "r");

	    if (infp)
	    {
		printf("Registering %s as nonspam.\n", arg1);
		SHIP("UNSPAM\n");
		ship(infp, fd);
		SHIP(".\n");
	    }
	}
	else if (sscanf(sendline, "u %s\n", arg1) == 1)
	{
	    FILE	*infp = fopen(arg1, "r");

	    if (infp)
	    {
		printf("Checking spamicity of %s.\n", arg1);
		SHIP("UNKNOWN\n");
		ship(infp, fd);
		SHIP(".\n");
		accept_response(fd, 2);
	    }
	}
	else if (sscanf(sendline, "w %s\n", arg1) == 1)
	{
		printf("Checking spamicity of %s.\n", arg1);
		sockprintf(fd, "WORD %s\n", arg1);
		accept_response(fd, 2);
	}
	else
	{
	    printf("Command summary:\n");
	    printf("s <file>       -- register contents of file as spam\n");
	    printf("n <file>       -- register contents of file as nonspam\n");
	    printf("S <file>       -- change contents of file to spam\n");
	    printf("N <file>       -- change contents of file to nonspam\n");
	    printf("u <file>       -- check the spamicity of file\n");
	    printf("w <word>       -- return good and bad counts for word\n");
	}
    }
  
    if (ferror(stdin))
	fprintf(stderr, "debug: error reading stdin\n");
    fprintf(stderr, "debug: signing off (%d)\n", getpid());
    return(0);
#undef SHIP
}

static int bogoserver(int fd, int initstat)
{
    byte_stuffed = 1;
#define ECHO(str)	sockputs(fd, str, strlen(str))
    for (;;)
    {
	char	cmdbuffer[BUFSIZ], word[BUFSIZ];

	if (sockgets(fd, cmdbuffer, sizeof(cmdbuffer)) <= 0)
	    break;
	if (initstat)
	    ECHO("ERROR\n");
	else if (!strcmp("SPAM\n", cmdbuffer))
	{
	    register_words(fd, &bad_list, NULL, -1);
	    write_lists();
	}
	else if (!strcmp("UNSPAM\n", cmdbuffer))
	{
	    register_words(fd, &good_list, &bad_list, -1);
	    write_lists();
	}
	else if (!strcmp("NONSPAM\n", cmdbuffer))
	{
	    register_words(fd, &good_list, NULL, -1);
	    write_lists();
	}
	else if (!strcmp("ENSPAM\n", cmdbuffer))
	{
	    register_words(fd, &bad_list, &good_list, -1);
	    write_lists();
	}
	else if (!strcmp("UNKNOWN\n", cmdbuffer))
	{
	    bogostat_t	*stats = bogofilter(fd);

	    statdump(stats, fd);
	    if (stats->spamicity <= SPAM_CUTOFF)
		sockprintf(fd, "NONSPAM\n");
	    else
		sockprintf(fd, "SPAM\n");
	}
	else if (sscanf(cmdbuffer, "WORD %s\n", word) == 1)
	{
	    sockprintf(fd, "# Goodness = %d, badness=%d\n", 
		 getcount(word, &good_list), getcount(word, &bad_list));
	    sockprintf(fd, "OK\n");
	}
	else
	    ECHO("EH?\n");
    }
#undef ECHO
}

enum mode_t {UNKNOWN, SPAM, NONSPAM_TO_SPAM, NONSPAM, SPAM_TO_NONSPAM, DEBUG,};

static int (*clienthook)(int fd) = unknown;
static int mode = UNKNOWN;

static int bogoinit()
// initialization hook for bogofilter server
{
    int readerror = 0;

    readerror += read_list(goodfile, &good_list);
    readerror += read_list(badfile,  &bad_list);

    /* readerror is ok, but only if using register_* modes */
    if (readerror && ! (clienthook == register_spam || clienthook == register_nonspam)) {
	// log("Error: can't open list file(s).\n");
	return(-1);
    }
    else
	return(0);
}

int main(int argc, char **argv)
{
    int	ch;
    struct stat sb;
    char dir[PATH_MAX];
    extern char *optarg;
    extern int optind, opterr, optopt;
    int foreground = 0;

    strcpy(dir, getenv("HOME"));
    strcat(dir, BOGODIR);

    while ((ch = getopt(argc, argv, "d:FnNpsSxv")) != EOF)
	switch(ch)
	{
	case 'd':
	    strcpy(dir, optarg);
	    break;
	case 'F':
	    foreground = 1;
	    break;
        case 'p':
	    passthrough = 1;
	case 'S':
	    clienthook = nonspam_to_spam;
	    mode = NONSPAM_TO_SPAM;
	    break;
	case 's':
	    clienthook = register_spam;
	    mode = SPAM;
	    break;
	case 'N':
	    clienthook = nonspam_to_spam;
	    mode = SPAM_TO_NONSPAM;
	    break;
	case 'n':
	    clienthook = register_nonspam;
	    mode = NONSPAM;
	    break;
	case 'x':
	    clienthook = debug;
	    mode = DEBUG;
	    break;
	case 'v':
	    verbose++;
	    break;
	}

    ch = stat(dir, &sb);
    if (!((ch == 0 && S_ISDIR(sb.st_mode))
	  || (errno==ENOENT && mkdir(dir, S_IRUSR|S_IWUSR|S_IXUSR)==0)))
    {
	fprintf(stderr,"bogofilter: something is wrong with %s.\n", dir);
	exit(2);
    }

    strcpy(goodfile, dir);
    strcat(goodfile, GOODFILE);
    strcpy(badfile, dir);
    strcat(badfile, BADFILE);

    if (!foreground)
    {
	int status = autodaemon("bogofilter", bogoinit,clienthook,bogoserver);

	if (passthrough)
	    for (textend=&textblocks; textend->block; textend=textend->next)
	    {
		int	inheaders = 1;

		if (inheaders && strncmp(textend->block, "Subject:", 8)==0)
		    printf("X-Spam-Status: %s tests=bogofilter\n", 
			   status ? "No" : "Yes");
		else if (strcmp(textend->block, "\n") == 0 && inheaders == 0)
		    inheaders = 0;
		(void) fputs(textend->block, stdout);
	    }

	exit(status);
    }
    else
    {
	bogoinit();
	switch(mode)
	{
	case SPAM:
	    register_words(0, &bad_list, NULL, 1);
	    write_lists();
	    break;
	case NONSPAM_TO_SPAM:
	    register_words(0, &good_list, &bad_list, 1);
	    write_lists();
	    break;
	case NONSPAM:
	    register_words(0, &good_list, NULL, 1);
	    write_lists();
	    break;
	case SPAM_TO_NONSPAM:
	    register_words(0, &bad_list, &good_list, 1);
	    write_lists();
	    break;
	case UNKNOWN:
	    {
		bogostat_t	*stats = bogofilter(0);

		statdump(stats, 1);
		exit(stats->spamicity <= SPAM_CUTOFF? 1 : 0);
	    }
	case DEBUG:
	    fprintf(stderr,
		    "bogofilter: debugger not available in foreground\n");
	    exit(1);
	}
    }
    exit(0);
}

// End
